package MakeRedirect;
use strict;
use warnings;
use MakeRedirect::URI;
use Text::CSV;
use Text::Locus;
use Carp;

our $VERSION = '0.10';

sub new {
    my ($class, $output, @args) = @_;
    my $self = bless {}, $class;
    my $modname = __PACKAGE__ . '::Output::' . $output;
    my $modpath = $modname;
    $modpath =~ s{::}{/}g;
    $modpath .= '.pm';
    $self->{output} = eval { require $modpath; $modname->new(@args) };

    if ($@) {
        if ($@ =~ /Can't locate $modpath/) {
	    croak "unknown output module: $output\n"
        }
        croak $@;
    }
    
    return $self;
}

sub scheme {
    my ($self, $scheme) = @_;
    if ($scheme) {
	$self->{scheme} = $scheme;
    }
    $self->{scheme};
}

sub getopt {
    my $self = shift;
    $self->{output}->getopt;
}

sub output {
    my $self = shift;
    $self->{output}->open;
    foreach my $host (sort keys %{$self->{rules}}) {
	$self->{output}->ruleset($host, $self->{rules}{$host});
    }
    $self->{output}->close;
}

sub error {
    my $self = shift;
    warn "@_\n";
}

sub read {
    my ($self, $file, %opts) = @_;

    my $skiplines = delete $opts{skip};
    my $filename = delete $opts{filename};
    my $fh;
    if (ref($file) eq 'GLOB') {
	open($fh, '<&', $file)
	    or croak "can't dup file handle: $!";
	$filename = '<filehandle>' unless $filename;
    } else {
	open($fh, '<:encoding(utf8)', $file)
	    or croak "Can't open $file: $!";
	$filename = $file unless $filename;
    }
    
    use constant {
	SRC => 0,
	DST => 1,
	NF  => 2
    };

    my $csv = Text::CSV->new( { binary => 1 } )
	or croak "Cannot use CSV: ".Text::CSV->error_diag();
    my $line = 0;
    while (my $row = $csv->getline($fh)) {
	$line++;
        next if $skiplines && $line <= $skiplines;
        my $nf = @$row;
        if ($nf == 0) {
	    next
        }
        if ($nf < NF) {
	    $self->error("$filename:$line: too few fields; ignoring line");
	    next;
        }
        if ($nf > NF) {
	    $self->error("$filename:$line: too many fields; ignoring surplus ones");
	    next;
        }

	my $src = new MakeRedirect::URI($row->[SRC],
					scheme => $self->{scheme});
	my $dst = new MakeRedirect::URI($row->[DST],
					scheme => $self->{scheme});
        if ($src == $dst) {
	    $self->error("$filename:$line: source and destination are the same");
	    next;
        }

	my $host = $src->host // '';
	my $locus = new Text::Locus($filename, $line);
	if (my $orig = $self->{rules}{$host}{$src->canonical}) {
	    if ($orig->dst == $dst) {
		$self->error("$locus: duplicate rule (original declaration at "
		             . $orig->locus . "); ignoring");
	        next;
	    } else {
		$self->error("$locus: redefinition of the rule at ("
		             . $orig->locus . ")\n");
	    }
	}
	$self->{rules}{$host}{$src->canonical} =
	    MakeRedirect::Rule->new($src, $dst, $locus);
    }
    $csv->eof or $self->error($csv->error_diag());
    close $fh;
}


package MakeRedirect::Rule;
sub new {
    my ($class, $src, $dst, $locus) = @_;
    bless { src => $src, dst => $dst, locus => $locus }, $class
}

sub src { shift->{src} }
sub dst { shift->{dst} }
sub locus { shift->{locus} }

1;    
